from __future__ import print_function
import os, sys
import argparse
import warnings
import numpy as np
from matplotlib import colors
import matplotlib.pyplot as plt
from sklearn.gaussian_process import GaussianProcessRegressor
from sklearn.gaussian_process.kernels import RBF, ConstantKernel as C

### set argparse

parser = argparse.ArgumentParser(description="This software evaluates the accuracy of a profile-based predictor as a function of the max cosine similarity between the training and test profiles. See input/output file formats in the README.txt file.",
epilog="Ex: python "+sys.argv[0]+" --sets test/PSSM_train.dat test/PSSM_test.dat --out TEST")
parser.add_argument("--sets", metavar='file', type=str, nargs=2, help="training and test datasets, in that order [REQUIRED]")
parser.add_argument("--out", metavar="output_prefix", type=str, help="prefix for the name of the output files [REQUIRED]")
parser.add_argument("--step", metavar="window_size", type=int, default=30, help="fixed length of the profile windows [DEFAULT=30]")
parser.add_argument("--cpu", metavar="num_thread", type=int, default=1, help="number of threads to use for the analysis [DEFAULT=1]")
parser.add_argument("--keep", action="store_true", help="keep full results in separate output file [OPTIONAL]")
parser.add_argument("--warn", action="store_true", help="display all warning messages on screen [OPTIONAL]")
args = parser.parse_args()

# check arguments

if not (args.sets and args.out):
    print('\n---->> You need to specify the paths of the input datasets and the output file name prefix.\nEx: python',sys.argv[0],'--sets test/PSSM_train.dat test/PSSM_test.dat --out TEST\nCheck the README.txt file for more details or type: python',sys.argv[0],'--help')
    exit()

cwd = os.getcwd()
wd = os.path.abspath(os.path.dirname(sys.argv[0]))+"/lib"
training = os.path.abspath(args.sets[0])
test = os.path.abspath(args.sets[1])
raw_out = os.path.abspath(args.out)+".raw"
os.chdir(wd)

if not os.path.isfile('multicore_analysis'):
    os.system('gcc -Wall multicore_analysis.c -o multicore_analysis -lm')
    os.chmod('multicore_analysis', 0o755)
os.system('./multicore_analysis.pl %s %s %s %d %d' % (training, test, raw_out, args.step, args.cpu))
os.chdir(cwd)

if not args.warn:
    def warn(*args, **kwargs):
        pass
    warnings.warn = warn

# GPR

x = []
y = []
with open(raw_out, "r") as preprocessed:
    preprocessed.readline()
    for line in preprocessed:
        splitted = line.split("\t")
        x.append(float(splitted[2]))
        y.append(float(splitted[3]))
x = np.asarray(x)
y = np.asarray(y)


def gregs_mean_average_val(X, Y, num_buckets):
    '''Returns: x, y, counts
    '''
    mi, ma = min(X), max(X)
    x = np.linspace(mi, ma, num_buckets)
    y = np.zeros((num_buckets,), 'float32')
    counts = np.zeros((num_buckets,), 'float32')

    bucket_size = (ma - mi)*1./num_buckets + 1e-9
    for xx, yy in zip(X-mi, Y):
        pos = int(xx / bucket_size)
        counts[pos] += 1
        y[pos] += yy
    return x, y/(counts+1e-10), counts


def remove_nullcount_points(X, Y, counts):
    rx, ry, rc = [], [], []
    for x,y,c in zip(X, Y, counts):
        if c:
            rx.append(x)
            ry.append(y)
            rc.append(c)
    return map(np.asarray, [rx, ry, rc])


def smoothing(what, N=1):
    if N<1:
        return what
    filt = np.asarray([1,2,4,8,16,8,4,2,1], 'float32')
    filter_loss = (len(filt)-1)*N

    X = np.zeros(len(what)+filter_loss, 'float32')
    X[filter_loss//2:-filter_loss//2] = what
    X[-filter_loss//2:] = what[-filter_loss//2:][::-1] # mirror

    for _ in range(N):
        X = np.convolve(X, filt/filt.sum(), mode='valid')
    return X


def GPR_noise_aware(X, Y, guessed_noise_level = 0.06, newfigure=1, color='b', rel_bin_X=None, rel_bin_count=None, density_smoothing=0):
    np.random.seed(1)
    x = np.atleast_2d(np.linspace(min(X), max(X), 1000))
    x = x.reshape((len(x),1))
    dy = guessed_noise_level * np.ones(Y.shape)
    kernel = C(1.0, (0.1, 10)) * RBF(10, (1e-2, 1e2))
    # Instantiate a Gaussian Process model
    gp = GaussianProcessRegressor(kernel=kernel, alpha=dy ** 2,
                                  n_restarts_optimizer=10)
    # Fit to data using Maximum Likelihood Estimation of the parameters
    gp.fit(X, Y)
    # Make the prediction on the meshed x-axis (ask for MSE as well)
    y_pred, sigma = gp.predict(x, return_std=True)
    if newfigure:
        plt.figure()
    if 0:
        plt.errorbar(X.ravel(), Y, dy, fmt=color+'.', markersize=4, alpha=0.2)
    plt.plot(x, y_pred, color+'-')
    plt.fill(np.concatenate([x, x[::-1]]), np.concatenate([y_pred - 1.9600 * sigma, (y_pred + 1.9600 * sigma)[::-1]]),
             alpha=0.2, fc=color, ec='None')
    if newfigure:
        plt.xlabel('Cosine Similarity')
        plt.ylabel('Accuracy')
        plt.ylim(0, 1)

    if rel_bin_count is not None:
        x = np.atleast_2d(np.linspace(0,1, 1000)).T
        density = np.zeros(len(x), 'float32')
        for lx, ly in zip(rel_bin_X, rel_bin_count):
            density[min(len(x)-1, int(round(lx*len(x))))] += ly

        density = smoothing(density, N=density_smoothing)
        density = density * 0.5 / density.max()
        plt.fill(np.concatenate([x, x[::-1]]), np.concatenate([density, density*0]),
                 alpha=0.5, fc=color, ec='None')

    plt.xlim(min(X),1)
    #plt.legend(loc='center right', prop={'size': 9})
    return gp


def preprocess_data(X, Y, target_num_samples = 400):
    n_bins = target_num_samples * 5
    first_x, first_y, first_counts = remove_nullcount_points(*gregs_mean_average_val(list(X), list(Y), int(n_bins)))
    counts = first_counts
    x = first_x
    y = first_y
    NUM = len(x)
    while NUM > 400:
        n_bins = max(0.3 * n_bins, min(0.8 * n_bins, target_num_samples**2*1.1/NUM ))
        x, y, counts = remove_nullcount_points(*gregs_mean_average_val(list(X), list(Y), int(n_bins)))
        NUM = len(x)

    rel_importance = np.asarray(counts, 'float32')/max(counts)
    guessed_noise_level = 0.07 * (max(rel_importance)-rel_importance)
    return x, y, guessed_noise_level, (first_x, first_counts)

plt.switch_backend('agg')
x, y, guessed_noise_level, rel_bin_count = preprocess_data(x, y)
model = GPR_noise_aware(x[:,None], y, guessed_noise_level = guessed_noise_level, newfigure=True,
                        color='b', rel_bin_X=rel_bin_count[0], rel_bin_count=rel_bin_count[1], density_smoothing=5)
plt.savefig(os.path.abspath(args.out)+'.png'.format(5), dpi=900)
#plt.show()

# generates a simple excel/csv file containing any similarity value observed in the sets (using a 0.01 step): 
# + the predicted accuracy from the GPR-learned function.
with open(os.path.abspath(args.out)+".csv", "w") as postprocessed:
    min_x = round(min(x[:,None]),2)
    max_x = 1.001
    xx = np.atleast_2d(np.arange(min_x, max_x, 0.01))
    xx = xx.reshape(len(xx[0]),1)
    yy = model.predict(xx)
    postprocessed.write("Cosine Similarity\tAccuracy\n")
    for i, val in enumerate(yy):
        postprocessed.write(str('%.2f'%(i*0.01+min_x))+"\t"+str('%.4f'%val)+"\n")

# keep the list of all actual pairs cos sim / accuracy
if not args.keep:
    os.remove(raw_out)
print('\n---->>Program successfully executed.\n')

